"""

This is the documentation of the Overcooked Simulator.

# About the package

The package contains an environment for cooperation between players/agents. A PyGameGUI visualizes the game to
human or virtual agents in 2D. A 3D web-enabled version (for example for online studies, currently under development)
can be found [here](https://gitlab.ub.uni-bielefeld.de/scs/cocosy/godot-overcooked-3d-visualization).

# Background / Literature
The overcooked/cooking domain is a well established cooperation domain/task. There exists
environments designed for reinforcement learning agents as well as the game and adaptations of the game for human
players in a more "real-time" environment. They all mostly differ in the visual and graphics dimension. 2D versions
like overcooked-ai, ... are most known in the community. But more visual appealing 3D versions for cooperation with
humans are getting developed more frequently (cite,...). Besides, the general adaptations of the original overcooked
game.
With this overcooked-simulator, we want to bring both worlds together: the reinforcement learning and real-time playable
environment with an appealing visualisation. Enable the potential of developing artificial agents that play with humans
like a "real" cooperative / human partner.

# Installation

You need a Python **3.10** or newer environment.
```bash
pip install overcooked-environment@git+https://gitlab.ub.uni-bielefeld.de/scs/cocosy/overcooked-simulator@main
```
Or clone it and install it as an editable library which allows you to use all the scripts directly.
```bash
git clone https://gitlab.ub.uni-bielefeld.de/scs/cocosy/overcooked-simulator.git
cd overcooked_simulator
pip install -e .
```

# Usage / Examples
 Our overcooked simulator is designed for real time interaction but also for reinforcement
learning (gymnasium environment) with time independent step function calls. It focuses on configurability, extensibility and appealing visualization
options.

## Human Player
Run it via the command line (in your pyenv/conda environment):

```bash
overcooked-sim  --url "localhost" --port 8000
```

*The arguments shown are the defaults.*

You can also start the **Game Server** and the **PyGame GUI** individually in different terminals.

```bash
python3 overcooked_simulator/game_server.py --url "localhost" --port 8000 --manager_ids SECRETKEY1 SECRETKEY2

python3 overcooked_simulator/gui_2d_vis/overcooked_gui.py --url "localhost" --port 8000 --manager_ids SECRETKEY1
```

## Connect with agent and receive game state
Or you start a game server, create an environment and connect each player/agent via a websocket connection.

To start a game server see above. Your own manager needs to create an environment.
```python
import requests

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.game_server import CreateEnvironmentConfig
from overcooked_simulator.server_results import CreateEnvResult


with open(ROOT_DIR / "game_content" / "item_info.yaml", "r") as file:
    item_info = file.read()
with open(ROOT_DIR / "game_content" / "layouts" / "basic.layout", "r") as file:
    layout = file.read()
with open(ROOT_DIR / "game_content" / "environment_config.yaml", "r") as file:
    environment_config = file.read()

create_env = CreateEnvironmentConfig(
    manager_id="SECRETKEY1",
    number_players=2,
    environment_settings={"all_player_can_pause_game": False},
    item_info_config=item_info,
    environment_config=environment_config,
    layout_config=layout,
).model_dump(mode="json")

env_info = requests.post("http://localhost:8000/manage/create_env", json=create_env)
if env_info.status_code == 403:
    raise ValueError(f"Forbidden Request: {env_info.json()['detail']}")
env_info: CreateEnvResult = env_info.json()
```

Connect each player via a websocket (threaded or async).
```python
import json
import dataclasses
from websockets import connect

from overcooked_simulator.overcooked_environment import Action, ActionType, InterActionData
from overcooked_simulator.utils import custom_asdict_factory


p1_websocket = connect("ws://localhost:8000/ws/player/" + env_info["player_info"]["0"]["client_id"])

# set player "0" as ready
p1_websocket.send(json.dumps({"type": "ready", "player_hash": env_info["player_info"]["0"]["player_hash"]}))
assert json.loads(websocket.recv())["status"] == 200, "not accepted player"


# get the state for player "0", call every frame/step
p1_websocket.send(json.dumps({"type": "get_state", "player_hash": env_info["player_info"]["0"]["player_hash"]}))
state = json.loads(websocket.recv())

# send an action for player "0"
# --- movement ---
action = Action(
    player="0",
    action_type=ActionType.MOVEMENT,
    action_data=[0.0, 1.0],  # direction (here straight up)
    duration=0.5  # seconds
)
# --- pickup/drop off ---
action = Action(
    player="0",
    action_type=ActionType.PUT,
    action_data="pickup",
)
# --- interact ---
action = Action(
    player="0",
    action_type=ActionType.INTERACT,
    action_data=InterActionData.START  # InterActionData.STOP when to stop the interaction
)

p1_websocket.send(json.dumps({
    "type": "action",
    "player_hash": env_info["player_info"]["0"]["player_hash"],
    "action": dataclasses.asdict(
        action, dict_factory=custom_asdict_factory
    ),
}))
websocket.recv()

```

Stop the environment if you want the game to end before the time is up.
```python
requests.post(
    "http://localhost:8000/manage/stop_env",
    json={
        "manager_id": "SECRETKEY1",
        "env_id": env_info["env_id"],
        "reason": "closed environment",
    },
)
```

## Direct integration into your code
You can use the `overcooked_simulator.overcooked_environment.Environment` class
and call the `step`, `get_json_state`, and `perform_action` methods directly.

```python
from datetime import timedelta

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.overcooked_environment import Action, Environment

env = Environment(
    env_config=ROOT_DIR / "game_content" / "environment_config.yaml",
    layout_config=ROOT_DIR / "game_content" / "layouts" / "basic.layout",
    item_info=ROOT_DIR / "game_content" / "item_info.yaml"
)

env.add_player("0")
env.add_player("1")

while True:
    # adapt this to real time if needed with time.sleep etc.
    env.step(timedelta(seconds=0.1))

    player_0_state = json.loads(env.get_json_state("0"))
    if player_0_state["ended"]:
        break

    action = ...  # See above but use np.array instead of list for the movement direction vector
    env.perform_action(action)
```

# JSON State
The JSON Schema for the state of the environment for a player can be generated by running the `state_representation.py`
```bash
python state_representation.py
```
Should look like
```
{'$defs': {'CookingEquipmentState': {'properties': {'content_list': {'items': {'$ref': '#/$defs/ItemState'}, 'title': 'Content List', 'type': 'array'}, 'content_ready': {'anyOf': [{'$ref': '#/$defs/ItemState'}, {'type': 'null'}]}}, 'required': ['content_list', 'content_ready'], 'title': 'CookingEquipmentState', 'type': 'object'}, 'CounterState': {'properties': {'id': {'title': 'Id', 'type': 'string'}, 'category': {'const': 'Counter', 'title': 'Category'}, 'type': {'title': 'Type', 'type': 'string'}, 'pos': {'items': {'type': 'number'}, 'title': 'Pos', 'type': 'array'}, 'occupied_by': {'anyOf': [{'items': {'anyOf': [{'$ref': '#/$defs/ItemState'}, {'$ref': '#/$defs/CookingEquipmentState'}]}, 'type': 'array'}, {'$ref': '#/$defs/ItemState'}, {'$ref': '#/$defs/CookingEquipmentState'}, {'type': 'null'}], 'title': 'Occupied By'}}, 'required': ['id', 'category', 'type', 'pos', 'occupied_by'], 'title': 'CounterState', 'type': 'object'}, 'ItemState': {'properties': {'id': {'title': 'Id', 'type': 'string'}, 'category': {'anyOf': [{'const': 'Item'}, {'const': 'ItemCookingEquipment'}], 'title': 'Category'}, 'type': {'title': 'Type', 'type': 'string'}, 'progress_percentage': {'anyOf': [{'type': 'number'}, {'type': 'integer'}], 'title': 'Progress Percentage'}}, 'required': ['id', 'category', 'type', 'progress_percentage'], 'title': 'ItemState', 'type': 'object'}, 'KitchenInfo': {'properties': {'width': {'title': 'Width', 'type': 'number'}, 'height': {'title': 'Height', 'type': 'number'}}, 'required': ['width', 'height'], 'title': 'KitchenInfo', 'type': 'object'}, 'OrderState': {'properties': {'id': {'title': 'Id', 'type': 'string'}, 'category': {'const': 'Order', 'title': 'Category'}, 'meal': {'title': 'Meal', 'type': 'string'}, 'start_time': {'format': 'date-time', 'title': 'Start Time', 'type': 'string'}, 'max_duration': {'title': 'Max Duration', 'type': 'number'}}, 'required': ['id', 'category', 'meal', 'start_time', 'max_duration'], 'title': 'OrderState', 'type': 'object'}, 'PlayerState': {'properties': {'id': {'title': 'Id', 'type': 'string'}, 'pos': {'items': {'type': 'number'}, 'title': 'Pos', 'type': 'array'}, 'facing_direction': {'items': {'type': 'number'}, 'title': 'Facing Direction', 'type': 'array'}, 'holding': {'anyOf': [{'$ref': '#/$defs/ItemState'}, {'$ref': '#/$defs/CookingEquipmentState'}, {'type': 'null'}], 'title': 'Holding'}, 'current_nearest_counter_pos': {'anyOf': [{'items': {'type': 'number'}, 'type': 'array'}, {'type': 'null'}], 'title': 'Current Nearest Counter Pos'}, 'current_nearest_counter_id': {'anyOf': [{'type': 'string'}, {'type': 'null'}], 'title': 'Current Nearest Counter Id'}}, 'required': ['id', 'pos', 'facing_direction', 'holding', 'current_nearest_counter_pos', 'current_nearest_counter_id'], 'title': 'PlayerState', 'type': 'object'}}, 'properties': {'players': {'items': {'$ref': '#/$defs/PlayerState'}, 'title': 'Players', 'type': 'array'}, 'counters': {'items': {'$ref': '#/$defs/CounterState'}, 'title': 'Counters', 'type': 'array'}, 'kitchen': {'$ref': '#/$defs/KitchenInfo'}, 'score': {'anyOf': [{'type': 'number'}, {'type': 'integer'}], 'title': 'Score'}, 'orders': {'items': {'$ref': '#/$defs/OrderState'}, 'title': 'Orders', 'type': 'array'}, 'ended': {'title': 'Ended', 'type': 'boolean'}, 'env_time': {'format': 'date-time', 'title': 'Env Time', 'type': 'string'}, 'remaining_time': {'title': 'Remaining Time', 'type': 'number'}}, 'required': ['players', 'counters', 'kitchen', 'score', 'orders', 'ended', 'env_time', 'remaining_time'], 'title': 'StateRepresentation', 'type': 'object'}
```

The BaseModel and TypedDicts can be found in `overcooked_simulator.state_representation`. The
`overcooked_simulator.state_representation.StateRepresentation` represents the json state that the `get_json_state`
returns.

## Generate images from JSON states
You might have stored some json states and now you want to visualize them via the
pygame-2d visualization. You can do that by running the `drawing.py` script and referencing a json file.
```bash
python3 overcooked_simulator/gui_2d_vis/drawing.py --state my_state.json
```
- You can specify a different visualization config with `-v` or `--visualization_config`.
- You can specify the name of the output file with `-o` or `--output_file`. The default is `screenshot.jpg`.

# Configuration

The environment configuration is currently done with 3 config files + GUI configuration.

## Item Config

The item config defines which ingredients, cooking equipment and meals can exist and how meals and processed ingredients
can be cooked/created.

For example

    CuttingBoard:
      type: Equipment

    Stove:
      type: Equipment

    Pot:
      type: Equipment
      equipment: Stove

    Tomato:
      type: Ingredient

    ChoppedTomato:
      type: Ingredient
      needs: [ Tomato ]
      seconds: 4.0
      equipment: CuttingBoard

    TomatoSoup:
      type: Meal
      needs: [ ChoppedTomato, ChoppedTomato, ChoppedTomato ]
      seconds: 6.0
      equipment: Pot


## Layout Config

You can define the layout of the kitchen via a layout file. The position of counters are based on a grid system, even
when the players do not move grid steps but continuous steps. Each character defines a different type of counter. Which
character is mapped to which counter is defined in the Environment config.

For example

```
#QU#FO#TNLB#
#__________M
#__________K
W__________I
#__A_____A_D
C__________E
C__________G
#__________#
#P#S+#X##S+#
```

## Environment Config

The environment config defines how a level/environment is defined. Here, the available plates, meals, order and player
configuration is done.

For example

```yaml
plates:
  clean_plates: 1
  dirty_plates: 2
  plate_delay: [ 5, 10 ]
  # range of seconds until the dirty plate arrives.

game:
  time_limit_seconds: 300

meals:
  all: true

layout_chars:
  _: Free
  hash: Counter
  A: Agent
  P: PlateDispenser
  C: CuttingBoard
  X: Trashcan
  W: ServingWindow
  S: Sink
  +: SinkAddon
  U: Pot  # with Stove
  T: Tomato

orders:
  ...

player_config:
  radius: 0.4
  player_speed_units_per_seconds: 8
  interaction_range: 1.6
```

## PyGame Visualization Config

Here the visualisation for all objects is defined. Reference the images or define a list of base shapes that represent
the counters, ingredients, meals and players.


# Citation

# Structure of the Documentation
The API documentation follows the file and content structure in the repo.
On the left you can find the navigation panel that brings you to the implementation of
- the **counter factory** converts the characters in the layout file to counter instances,
- the **counters**, including the kitchen utility objects like dispenser, cooking counter (stove, deep fryer, oven),
  sink, etc.,
- the **game items**, the holdable ingredients, cooking equipment, composed ingredients, and meals,
- the **game server**, which can manage several running environments and can communicates via FastAPI post requests and
websockets,
- the **orders**, how to sample incoming orders and their attributes,
- the **environment**, handles the incoming actions and provides the state,
- the **player**/agent, that interacts in the environment,
- type hints are defined in **state representation** for the json state and **server results** for the data returned by
the game server in post requests.
- **util**ity code.


"""
import os
from pathlib import Path

ROOT_DIR = Path(os.path.dirname(os.path.abspath(__file__)))  # This is your Project Root
"""A path variable to get access to the layouts coming with the package. For example,
```python 
from overcooked_simulator import ROOT_DIR

environment_config_path = ROOT_DIR / "game_content" / "environment_config.yaml"
```
"""
